@@ -30,9 +30,9 @@
 #include "RosterDb.h"
 #include "RosterModel.h"
 
-bool DisplayedMessageReaction::operator<(const DisplayedMessageReaction &other) const
+auto DisplayedMessageReaction::operator<=>(const DisplayedMessageReaction &other) const
 {
-    return emoji < other.emoji;
+    return emoji <=> other.emoji;
 }
 
 bool DetailedMessageReaction::operator<(const DetailedMessageReaction &other) const
@@ -293,27 +293,25 @@ QVariant MessageModel::data(const QModelIndex &index, int role) const
         QList<DisplayedMessageReaction> displayedMessageReactions;
 
         const auto &reactionSenders = msg.reactionSenders;
-        for (auto itr = reactionSenders.begin(); itr != reactionSenders.end(); ++itr) {
+        for (auto itr = reactionSenders.cbegin(); itr != reactionSenders.cend(); ++itr) {
             const auto ownReactionsIterated = itr.key() == m_accountSettings->jid();
 
             for (const auto &reaction : std::as_const(itr->reactions)) {
-                auto reactionItr = std::find_if(displayedMessageReactions.begin(),
-                                                displayedMessageReactions.end(),
-                                                [=](const DisplayedMessageReaction &displayedMessageReaction) {
-                                                    return displayedMessageReaction.emoji == reaction.emoji;
-                                                });
+                auto reactionItr = std::ranges::find_if(displayedMessageReactions, [=](const DisplayedMessageReaction &displayedMessageReaction) {
+                    return displayedMessageReaction.emoji == reaction.emoji;
+                });
 
                 if (ownReactionsIterated) {
                     if (reactionItr == displayedMessageReactions.end()) {
-                        displayedMessageReactions.append({reaction.emoji, 1, ownReactionsIterated, reaction.deliveryState});
+                        displayedMessageReactions.append({reaction.emoji, 1, true, reaction.deliveryState});
                     } else {
                         reactionItr->count++;
-                        reactionItr->ownReactionIncluded = ownReactionsIterated;
+                        reactionItr->ownReactionIncluded = true;
                         reactionItr->deliveryState = reaction.deliveryState;
                     }
                 } else {
                     if (reactionItr == displayedMessageReactions.end()) {
-                        displayedMessageReactions.append({reaction.emoji, 1, ownReactionsIterated, {}});
+                        displayedMessageReactions.append({reaction.emoji, 1, false, {}});
                     } else {
                         reactionItr->count++;
                     }
@@ -339,7 +337,7 @@ QVariant MessageModel::data(const QModelIndex &index, int role) const
                         emojis.append(reaction.emoji);
                     }
 
-                    std::sort(emojis.begin(), emojis.end());
+                    std::ranges::sort(emojis);
 
                     detailedMessageReactions.append({itr.key(), {}, {}, emojis});
                 }
@@ -352,7 +350,7 @@ QVariant MessageModel::data(const QModelIndex &index, int role) const
     }
     case OwnReactionsFailed: {
         const auto &ownReactions = msg.reactionSenders.value(m_accountSettings->jid()).reactions;
-        return std::any_of(ownReactions.cbegin(), ownReactions.cend(), [](const MessageReaction &reaction) {
+        return std::ranges::any_of(ownReactions, [](const MessageReaction &reaction) {
             switch (reaction.deliveryState) {
             case MessageReactionDeliveryState::ErrorOnAddition:
             case MessageReactionDeliveryState::ErrorOnRemovalAfterSent:
@@ -424,11 +422,11 @@ void MessageModel::fetchMore(const QModelIndex &)
             const auto isGroupChat = m_chatController->rosterItem().isGroupChat();
 
             if (m_messages.isEmpty()) {
-                m_messageController->retrieveBacklogMessages(chatJid, isGroupChat).then(this, [this](bool complete) {
+                m_messageController->retrieveBacklogMessages(chatJid, isGroupChat).then([this](bool complete) {
                     handleMamBacklogRetrieved(complete);
                 });
             } else {
-                m_messageController->retrieveBacklogMessages(chatJid, isGroupChat, m_messages.constLast().stanzaId).then(this, [this](bool complete) {
+                m_messageController->retrieveBacklogMessages(chatJid, isGroupChat, m_messages.constLast().stanzaId).then([this](bool complete) {
                     handleMamBacklogRetrieved(complete);
                 });
             }
@@ -443,6 +441,21 @@ bool MessageModel::canFetchMore(const QModelIndex &) const
     return !m_fetchedAllFromDb || (!m_fetchedAllFromMam && !m_mamLoading);
 }
 
+void MessageModel::resendMessage(int index)
+{
+    if (index < 0 || index >= m_messages.size()) {
+        return;
+    }
+
+    const auto message = m_messages.at(index);
+
+    MessageDb::instance()->updateMessage(m_accountSettings->jid(), m_chatController->jid(), message.relevantId(), [](Message &message) {
+        message.deliveryState = DeliveryState::Pending;
+    });
+
+    m_messageController->sendPendingMessage(std::move(message));
+}
+
 void MessageModel::handleMessageRead(int readMessageIndex)
 {
     // Check the index validity.
@@ -519,7 +532,7 @@ void MessageModel::setMessageMarked(int index, bool marked)
 
 void MessageModel::addMessageReaction(const QString &messageId, const QString &emoji)
 {
-    const auto itr = std::find_if(m_messages.cbegin(), m_messages.cend(), [&](const Message &message) {
+    const auto itr = std::ranges::find_if(m_messages, [&](const Message &message) {
         return message.relevantId() == messageId;
     });
 
@@ -576,27 +589,26 @@ void MessageModel::addMessageReaction(const QString &messageId, const QString &e
                                           emojis,
                                           m_chatController->activeEncryption(),
                                           m_chatController->groupChatUserJids())
-                    .then(this, [this, addReaction, messageId, senderId, emoji](QXmpp::SendResult &&result) {
+                    .then([this, addReaction, messageId, senderId, emoji](QXmpp::SendResult &&result) {
                         if (const auto error = std::get_if<QXmppError>(&result)) {
                             Q_EMIT MainController::instance()->passiveNotificationRequested(tr("Reaction could not be sent: %1").arg(error->description));
 
-                            MessageDb::instance()->updateMessage(
-                                m_chatController->account()->settings()->jid(),
-                                m_chatController->jid(),
-                                messageId,
-                                [senderId, emoji](Message &message) {
-                                    auto &reactionSender = message.reactionSenders[senderId];
-                                    reactionSender.latestTimestamp = QDateTime::currentDateTimeUtc();
-                                    auto &reactions = reactionSender.reactions;
+                            MessageDb::instance()->updateMessage(m_chatController->account()->settings()->jid(),
+                                                                 m_chatController->jid(),
+                                                                 messageId,
+                                                                 [senderId, emoji](Message &message) {
+                                                                     auto &reactionSender = message.reactionSenders[senderId];
+                                                                     reactionSender.latestTimestamp = QDateTime::currentDateTimeUtc();
+                                                                     auto &reactions = reactionSender.reactions;
 
-                                    auto itr = std::find_if(reactions.begin(), reactions.end(), [emoji](const MessageReaction &reaction) {
-                                        return reaction.emoji == emoji;
-                                    });
+                                                                     auto itr = std::ranges::find_if(reactions, [emoji](const MessageReaction &reaction) {
+                                                                         return reaction.emoji == emoji;
+                                                                     });
 
-                                    if (itr != reactions.end()) {
-                                        itr->deliveryState = MessageReactionDeliveryState::ErrorOnAddition;
-                                    }
-                                });
+                                                                     if (itr != reactions.end()) {
+                                                                         itr->deliveryState = MessageReactionDeliveryState::ErrorOnAddition;
+                                                                     }
+                                                                 });
                         } else {
                             m_messageController->updateMessageReactionsAfterSending(m_chatController->jid(), messageId, senderId);
                         }
@@ -608,7 +620,7 @@ void MessageModel::addMessageReaction(const QString &messageId, const QString &e
 
 void MessageModel::removeMessageReaction(const QString &messageId, const QString &emoji)
 {
-    const auto itr = std::find_if(m_messages.cbegin(), m_messages.cend(), [&](const Message &message) {
+    const auto itr = std::ranges::find_if(m_messages, [&](const Message &message) {
         return message.relevantId() == messageId;
     });
 
@@ -628,10 +640,9 @@ void MessageModel::removeMessageReaction(const QString &messageId, const QString
                                                                auto &reactionSenders = message.reactionSenders;
                                                                auto &reactions = reactionSenders[senderId].reactions;
 
-                                                               const auto itr =
-                                                                   std::find_if(reactions.begin(), reactions.end(), [&](const MessageReaction &reaction) {
-                                                                       return reaction.emoji == emoji;
-                                                                   });
+                                                               const auto itr = std::ranges::find_if(reactions, [&](const MessageReaction &reaction) {
+                                                                   return reaction.emoji == emoji;
+                                                               });
 
                                                                switch (auto &deliveryState = itr->deliveryState) {
                                                                case MessageReactionDeliveryState::Sent:
@@ -674,7 +685,7 @@ void MessageModel::removeMessageReaction(const QString &messageId, const QString
                                           emojis,
                                           m_chatController->activeEncryption(),
                                           m_chatController->groupChatUserJids())
-                    .then(this, [this, messageId, senderId, emoji](QXmpp::SendResult &&result) {
+                    .then([this, messageId, senderId, emoji](QXmpp::SendResult &&result) {
                         if (const auto error = std::get_if<QXmppError>(&result)) {
                             Q_EMIT MainController::instance()->passiveNotificationRequested(tr("Reaction could not be sent: %1").arg(error->description));
 
@@ -684,10 +695,9 @@ void MessageModel::removeMessageReaction(const QString &messageId, const QString
                                                                  [senderId, emoji](Message &message) {
                                                                      auto &reactions = message.reactionSenders[senderId].reactions;
 
-                                                                     const auto itr =
-                                                                         std::find_if(reactions.begin(), reactions.end(), [&](const MessageReaction &reaction) {
-                                                                             return reaction.emoji == emoji;
-                                                                         });
+                                                                     const auto itr = std::ranges::find_if(reactions, [&](const MessageReaction &reaction) {
+                                                                         return reaction.emoji == emoji;
+                                                                     });
 
                                                                      switch (auto &deliveryState = itr->deliveryState) {
                                                                      case MessageReactionDeliveryState::Sent:
@@ -711,7 +721,7 @@ void MessageModel::removeMessageReaction(const QString &messageId, const QString
 
 void MessageModel::resendMessageReactions(const QString &messageId)
 {
-    const auto itr = std::find_if(m_messages.cbegin(), m_messages.cend(), [&](const Message &message) {
+    const auto itr = std::ranges::find_if(m_messages, [&](const Message &message) {
         return message.relevantId() == messageId;
     });
 
@@ -759,7 +769,7 @@ void MessageModel::resendMessageReactions(const QString &messageId)
                                       emojis,
                                       m_chatController->activeEncryption(),
                                       m_chatController->groupChatUserJids())
-                .then(this, [this, messageId, senderId](QXmpp::SendResult &&result) {
+                .then([this, messageId, senderId](QXmpp::SendResult &&result) {
                     if (const auto error = std::get_if<QXmppError>(&result)) {
                         Q_EMIT MainController::instance()->passiveNotificationRequested(tr("Reactions could not be sent: %1").arg(error->description));
 
@@ -794,6 +804,32 @@ void MessageModel::resendMessageReactions(const QString &messageId)
     }
 }
 
+int MessageModel::nextCorrectableMessageIndex(int indexOffset) const
+{
+    for (int i = indexOffset; i < m_messages.size(); i++) {
+        if (canCorrectMessage(i)) {
+            return i;
+        }
+    }
+
+    return -1;
+}
+
+int MessageModel::previousCorrectableMessageIndex(int indexOffset) const
+{
+    if (indexOffset >= m_messages.size()) {
+        return -1;
+    }
+
+    for (int i = indexOffset; i >= 0; i--) {
+        if (canCorrectMessage(i)) {
+            return i;
+        }
+    }
+
+    return -1;
+}
+
 bool MessageModel::canCorrectMessage(int index) const
 {
     // The message must be loaded.
@@ -801,15 +837,16 @@ bool MessageModel::canCorrectMessage(int index) const
         return false;
     }
 
-    // The message must be an own one and not contain an error.
-    const auto &msg = m_messages.at(index);
-    if (!msg.isOwn || msg.deliveryState == Enums::DeliveryState::Error) {
+    const auto &message = m_messages.at(index);
+
+    if (!message.isOwn || message.groupChatInvitation || message.deliveryState == Enums::DeliveryState::Error
+        || RosterModel::instance()->item(message.accountJid, message.chatJid)->isDeletedGroupChat()) {
         return false;
     }
 
     // The message must not be too old.
     const auto timeThreshold = QDateTime::currentDateTimeUtc().addDays(-MAX_MESSAGE_CORRECTION_DAYS);
-    if (msg.timestamp < timeThreshold) {
+    if (message.timestamp < timeThreshold) {
         return false;
     }
 
@@ -821,17 +858,31 @@ bool MessageModel::canCorrectMessage(int index) const
     return true;
 }
 
+void MessageModel::deleteFile(int index, const File &file)
+{
+    MessageDb::instance()->updateMessage(m_accountSettings->jid(), m_chatController->jid(), m_messages.at(index).id, [fileId = file.id](Message &message) {
+        auto it = std::ranges::find_if(message.files, [fileId](const auto &file) {
+            return file.id == fileId;
+        });
+        if (it != message.files.end()) {
+            it->localFilePath.clear();
+        }
+    });
+
+    MediaUtils::deleteDownloadedFile(file.localFilePath);
+}
+
 void MessageModel::removeMessage(const QString &messageId)
 {
     const auto hasCorrectId = [&messageId](const Message &message) {
         return message.relevantId() == messageId;
     };
 
-    const auto itr = std::find_if(m_messages.cbegin(), m_messages.cend(), hasCorrectId);
+    const auto itr = std::ranges::find_if(m_messages, hasCorrectId);
 
     // Update the roster item of the current chat.
     if (itr != m_messages.cend()) {
-        int readMessageIndex = std::distance(m_messages.cbegin(), itr);
+        int readMessageIndex = std::ranges::distance(m_messages.cbegin(), itr);
 
         const QString &lastReadContactMessageId = m_chatController->rosterItem().lastReadContactMessageId;
         const QString &lastReadOwnMessageId = m_chatController->rosterItem().lastReadOwnMessageId;
@@ -864,9 +915,14 @@ void MessageModel::removeMessage(const QString &messageId)
             }
         }
 
-        // Remove the message from the database and model.
+        // Remove the message from the database/model and delete included files.
 
         MessageDb::instance()->removeMessage(itr->accountJid, itr->chatJid, messageId);
+
+        for (auto &file : itr->files) {
+            MediaUtils::deleteDownloadedFile(file.localFilePath);
+        }
+
         updateLastReadOwnMessageId();
 
         QModelIndex index = createIndex(readMessageIndex, 0);
@@ -983,10 +1039,7 @@ void MessageModel::handleMessagesFetched(const QList<Message> &msgs)
 
     beginInsertRows(QModelIndex(), rowCount(), rowCount() + msgs.size() - 1);
 
-    for (auto msg : msgs) {
-        m_messages << msg;
-        addVideoThumbnails(msg);
-    }
+    m_messages.append(msgs);
 
     updateLastReadOwnMessageId();
     updateFirstUnreadContactMessageIndex();
@@ -1032,28 +1085,17 @@ void MessageModel::handleMessageUpdated(Message message)
     }
 
     for (int i = 0; i < m_messages.size(); i++) {
-        const auto oldMessage = m_messages.at(i);
+        const auto &oldMessage = m_messages.at(i);
         const auto oldId = oldMessage.id;
         const auto oldReplaceId = oldMessage.replaceId;
 
         // The updated message can be either a normal message, a first message correction, an own
-        // reflected group chat message or a subsequent messsage correction.
+        // reflected group chat message or a subsequent message correction.
         if ((!oldId.isEmpty() && (oldId == message.id || oldId == message.replaceId || oldId == message.originId))
             || (!oldReplaceId.isEmpty() && oldReplaceId == message.replaceId)) {
-            const auto oldMessageTimestamp = m_messages.at(i).timestamp;
-
-            beginRemoveRows(QModelIndex(), i, i);
-            m_messages.removeAt(i);
-            endRemoveRows();
-
-            // Insert the message at its original position if the date is unchanged.
-            // Otherwise, move it to its new position.
-            if (!m_messages.isEmpty() && (message.timestamp == oldMessageTimestamp)) {
-                insertMessage(i, message);
-            } else {
-                addMessage(message);
-            }
-
+            m_messages.replace(i, message);
+            const auto modelIndex = index(i);
+            Q_EMIT dataChanged(modelIndex, modelIndex);
             break;
         }
     }
@@ -1071,7 +1113,7 @@ void MessageModel::handleDevicesChanged(QList<QString> jids)
                 continue;
             }
 
-            m_atmController->trustLevel(XMLNS_OMEMO_2, senderJid, itr->senderKey).then(this, [this, itr](QXmpp::TrustLevel trustLevel) {
+            m_atmController->trustLevel(XMLNS_OMEMO_2, senderJid, itr->senderKey).then([this, itr](QXmpp::TrustLevel trustLevel) {
                 // TODO: Search message again here because itr may not be up-to-date
 
                 itr->preciseTrustLevel = trustLevel;
@@ -1106,7 +1148,6 @@ void MessageModel::insertMessage(int idx, const Message &msg)
     endInsertRows();
 
     updateLastReadOwnMessageId();
-    addVideoThumbnails(msg);
 }
 
 void MessageModel::removeMessages(const QString &accountJid, const QString &chatJid)
@@ -1116,35 +1157,6 @@ void MessageModel::removeMessages(const QString &accountJid, const QString &chat
     }
 }
 
-void MessageModel::addVideoThumbnails(const Message &message)
-{
-    auto &files = message.files;
-    std::for_each(files.begin(), files.end(), [this, messageId = message.id](const File &file) {
-        if (const auto localFileUrl = file.localFileUrl(); file.type() == MessageType::MessageVideo && file.localFileUrl().isValid()) {
-            MediaUtils::generateThumbnail(localFileUrl, file.mimeTypeName(), VIDEO_THUMBNAIL_EDGE_PIXEL_COUNT)
-                .then(this, [this, fileId = file.fileId(), messageId](const QByteArray &thumbnail) mutable {
-                    auto itr = std::find_if(m_messages.begin(), m_messages.end(), [messageId](const Message &message) {
-                        return message.relevantId() == messageId;
-                    });
-
-                    if (itr != m_messages.cend()) {
-                        auto &files = itr->files;
-
-                        const auto fileItr = std::find_if(files.begin(), files.end(), [fileId, messageId](const File &queriedFile) {
-                            return queriedFile.fileId() == fileId;
-                        });
-
-                        if (fileItr != files.cend()) {
-                            fileItr->thumbnail = thumbnail;
-                            const auto modelIndex = index(std::distance(m_messages.begin(), itr));
-                            Q_EMIT dataChanged(modelIndex, modelIndex, {Files});
-                        }
-                    }
-                });
-        }
-    });
-}
-
 void MessageModel::updateLastReadOwnMessageId()
 {
     const auto formerLastReadOwnMessageId = m_lastReadOwnMessageId;
@@ -1202,7 +1214,7 @@ void MessageModel::emitMessagesUpdated(const QList<QString> &messageIds, Message
 
 bool MessageModel::undoMessageReactionRemoval(const QString &messageId, const QString &senderJid, const QString &emoji, const QList<MessageReaction> &reactions)
 {
-    const auto reactionItr = std::find_if(reactions.begin(), reactions.end(), [&](const MessageReaction &reaction) {
+    const auto reactionItr = std::ranges::find_if(reactions, [&](const MessageReaction &reaction) {
         return reaction.emoji == emoji;
     });
 
@@ -1213,7 +1225,7 @@ bool MessageModel::undoMessageReactionRemoval(const QString &messageId, const QS
                                              [senderJid, emoji](Message &message) {
                                                  auto &reactions = message.reactionSenders[senderJid].reactions;
 
-                                                 const auto itr = std::find_if(reactions.begin(), reactions.end(), [&](const MessageReaction &reaction) {
+                                                 const auto itr = std::ranges::find_if(reactions, [&](const MessageReaction &reaction) {
                                                      return reaction.emoji == emoji;
                                                  });
 
@@ -1242,7 +1254,7 @@ bool MessageModel::undoMessageReactionAddition(const QString &messageId,
                                                const QString &emoji,
                                                const QList<MessageReaction> &reactions)
 {
-    const auto reactionItr = std::find_if(reactions.begin(), reactions.end(), [&](const MessageReaction &reaction) {
+    const auto reactionItr = std::ranges::find_if(reactions, [&](const MessageReaction &reaction) {
         return reaction.emoji == emoji
             && (reaction.deliveryState == MessageReactionDeliveryState::PendingAddition
                 || reaction.deliveryState == MessageReactionDeliveryState::ErrorOnAddition);
@@ -1256,7 +1268,7 @@ bool MessageModel::undoMessageReactionAddition(const QString &messageId,
                                                  auto &reactionSenders = message.reactionSenders;
                                                  auto &reactions = reactionSenders[senderJid].reactions;
 
-                                                 const auto itr = std::find_if(reactions.begin(), reactions.end(), [&](const MessageReaction &reaction) {
+                                                 const auto itr = std::ranges::find_if(reactions, [&](const MessageReaction &reaction) {
                                                      return reaction.emoji == emoji;
                                                  });
 
